"""Classes and methods to describe contract Consumers."""
from .pact import Pact
from .provider import Provider


class Consumer(object):
    """
    A Pact consumer.

    Use this class to describe the service making requests to the provider and
    then use `has_pact_with` to create a contract with a specific service:

    >>> from pact import Consumer, Provider
    >>> consumer = Consumer('my-web-front-end')
    >>> consumer.has_pact_with(Provider('my-backend-serivce'))
    """

    def __init__(self, name, service_cls=Pact, tags=None,
                 tag_with_git_branch=False, version='0.0.0'):
        """
        Create the Consumer class.

        :param name: The name of this Consumer. This will be shown in the Pact
            when it is published.
        :type name: str
        :param service_cls: Pact, or a sub-class of it, to use when creating
            the contracts. This is useful when you have a custom URL or port
            for your mock service and want to use the same value on all of
            your contracts.
        :type service_cls: pact.Pact
        :param tags: A list of strings to use as tags to use when publishing
            to a pact broker. Defaults to None.
        :type tags: list
        :param tag_with_git_branch: A flag to determine whether to
            automatically tag a pact with the current git branch name.
            Defaults to False.
        :type tag_with_git_branch: bool
        :param version: The version of this Consumer. This will be used when
            publishing pacts to a pact broker. Defaults to '0.0.0'
        """
        self.name = name
        self.service_cls = service_cls
        self.tags = tags
        self.tag_with_git_branch = tag_with_git_branch
        self.version = version

    def has_pact_with(self, provider, host_name='localhost', port=1234,
                      log_dir=None, ssl=False, sslcert=None, sslkey=None,
                      cors=False, publish_to_broker=False,
                      broker_base_url=None, broker_username=None,
                      broker_password=None, broker_token=None, pact_dir=None,
                      specification_version='2.0.0',
                      file_write_mode='overwrite'):
        """
        Create a contract between the `provider` and this consumer.

        If you are running the Pact mock service in a non-default location,
        you can provide the host name and port here:

        >>> from pact import Consumer, Provider
        >>> consumer = Consumer('my-web-front-end')
        >>> consumer.has_pact_with(
        ...   Provider('my-backend-serivce'),
        ...   host_name='192.168.1.1',
        ...   port=8000)

        :param provider: The provider service for this contract.
        :type provider: pact.Provider
        :param host_name: An optional host name to use when contacting the
            Pact mock service. This will need to be the same host name used by
            your code under test to contact the mock service. It defaults to:
            `localhost`.
        :type host_name: str
        :param port: The TCP port to use when contacting the Pact mock service.
            This will need to tbe the same port used by your code under test
            to contact the mock service. It defaults to: 1234
        :type port: int
        :param log_dir: The directory where logs should be written. Defaults to
            the current directory.
        :type log_dir: str
        :param ssl: Flag to control the use of a self-signed SSL cert to run
            the server over HTTPS , defaults to False.
        :type ssl: bool
        :param sslcert: Path to a custom self-signed SSL cert file, 'ssl'
            option must be set to True to use this option. Defaults to None.
        :type sslcert: str
        :param sslkey: Path to a custom key and self-signed SSL cert key file,
            'ssl' option must be set to True to use this option.
            Defaults to None.
        :type sslkey: str
        :param cors: Allow CORS OPTION requests to be accepted,
            defaults to False.
        :type cors: bool
        :param publish_to_broker: Flag to control automatic publishing of
            pacts to a pact broker. Defaults to False.
        :type publish_to_broker: bool
        :param broker_base_url: URL of the pact broker that pacts will be
            published to. Defaults to None.
        :type broker_base_url: str
        :param broker_username: Username to use when connecting to the pact
            broker if authentication is required. Defaults to None.
        :type broker_username: str
        :param broker_password: Password to use when connecting to the pact
            broker if authentication is required. Defaults to None.
        :type broker_password: str
        :param broker_token: Authentication token to use when connecting to
            the pact broker. Defaults to None.
        :type broker_token: str
        :param pact_dir: Directory where the resulting pact files will be
            written. Defaults to the current directory.
        :type pact_dir: str
        :param specification_version: The Pact Specification version to use.
            Defaults to '2.0.0'.
        :type version: str
        :param file_write_mode: How the mock service should apply multiple
            calls to .verify(). Pass 'overwrite' to overwrite the generated
            JSON file on every call to .verify() or pass 'merge' to merge all
            interactions into the same JSON file. When using 'merge', make
            sure to delete any existing JSON file before calling .verify()
            for the first time. Defaults to 'overwrite'.
        :type version: str
        :return: A Pact object which you can use to define the specific
            interactions your code will have with the provider.
        :rtype: pact.Pact
        """
        if not isinstance(provider, (Provider,)):
            raise ValueError(
                'provider must be an instance of the Provider class.')

        return self.service_cls(
            broker_base_url=broker_base_url,
            broker_username=broker_username,
            broker_password=broker_password,
            broker_token=broker_token,
            consumer=self,
            provider=provider,
            host_name=host_name,
            port=port,
            log_dir=log_dir,
            ssl=ssl,
            sslcert=sslcert,
            sslkey=sslkey,
            cors=cors,
            pact_dir=pact_dir,
            publish_to_broker=publish_to_broker,
            specification_version=specification_version,
            file_write_mode=file_write_mode)



"""Contract Message Provider."""
import os
import time

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3 import Retry
from multiprocessing import Process
from .verifier import Verifier
from .http_proxy import run_proxy

import logging
logging.getLogger("urllib3").setLevel(logging.ERROR)


class MessageProvider(object):
    """
    A Pact message provider.

    provider = MessageProvider(
        message_providers = {
            "a document created successfully": handler
        },
        provider='DocumentService',
        pact_dir='pacts',
        version='3.0.0'
    )
    """

    def __init__(
        self,
        message_providers,
        provider,
        consumer,
        pact_dir=os.getcwd(),
        version="3.0.0",
        proxy_host='localhost',
        proxy_port='1234'
    ):
        """Create a Message Provider instance."""
        self.message_providers = message_providers
        self.provider = provider
        self.consumer = consumer
        self.version = version
        self.pact_dir = pact_dir
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self._process = None

    def _proxy_url(self):
        return f'http://{self.proxy_host}:{self.proxy_port}'

    def _pact_file(self):
        return f'{self.consumer}-{self.provider}.json'.lower().replace(' ', '_')

    def _setup_states(self):
        message_handlers = {}
        for key, handler in self.message_providers.items():
            message_handlers[f'{key}'] = handler()

        resp = requests.post(f'{self._proxy_url()}/setup',
                             verify=False,
                             json={"messageHandlers": message_handlers})
        assert resp.status_code == 201, resp.text
        return message_handlers

    def _wait_for_server_start(self):
        """
        Wait for the Http Proxy to be ready.

        :rtype: None
        :raises RuntimeError: If there is a problem starting the Http Proxy.
        """
        s = requests.Session()
        retries = Retry(total=9, backoff_factor=0.5)
        http_mount = 'http://'
        s.mount(http_mount, HTTPAdapter(max_retries=retries))
        resp = s.get(f'{self._proxy_url()}/ping', verify=False)
        if resp.status_code != 200:
            self._stop_proxy()
            raise RuntimeError(
                'There was a problem starting the proxy: %s', resp.text
            )

    def _wait_for_server_stop(self):
        """Wait for server to finish, or raise exception after timeout."""
        retry = 20
        while True:
            self._process.terminate()
            time.sleep(0.1)
            try:
                assert not self._process.is_alive()
                return 0
            except AssertionError:
                if retry == 0:
                    raise RuntimeError("Process timed out")
                retry -= 1

    def _start_proxy(self):
        self._process = Process(target=run_proxy, args=(), daemon=True)
        self._process.start()
        self._wait_for_server_start()
        self._setup_states()

    def _stop_proxy(self):
        """Stop the Http Proxy."""
        if isinstance(self._process, Process):
            self._wait_for_server_stop()

    def verify(self):
        """Verify pact files with executable verifier."""
        pact_files = f'{self.pact_dir}/{self._pact_file()}'
        verifier = Verifier(provider=self.provider,
                            provider_base_url=self._proxy_url())
        return_code, _ = verifier.verify_pacts(pact_files, verbose=False)
        assert (return_code == 0), f'Expected returned_code = 0, actual = {return_code}'

    def __enter__(self):
        """
        Enter a Python context.

        Sets up the Http Proxy.
        """
        self._start_proxy()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit a Python context.

        Return False to cascade the exception in context manager's body.
        Otherwise it will be supressed and the test will always pass.
        """
        if (exc_type, exc_val, exc_tb) != (None, None, None):
            if exc_type is not None:
                self._stop_proxy()
                return False
        self._stop_proxy()
        return True



"""Constant values for the pact-python package."""
import os
from os.path import join, dirname, normpath


def broker_client_exe():
    """Get the appropriate executable name for this platform."""
    if os.name == 'nt':
        return 'pact-broker.bat'
    else:
        return 'pact-broker'


def message_exe():
    """Get the appropriate executable name for this platform."""
    if os.name == 'nt':
        return 'pact-message.bat'
    else:
        return 'pact-message'


def mock_service_exe():
    """Get the appropriate executable name for this platform."""
    if os.name == 'nt':
        return 'pact-mock-service.bat'
    else:
        return 'pact-mock-service'


def provider_verifier_exe():
    """Get the appropriate provider executable name for this platform."""
    if os.name == 'nt':
        return 'pact-provider-verifier.bat'
    else:
        return 'pact-provider-verifier'


BROKER_CLIENT_PATH = normpath(join(
    dirname(__file__), 'bin', 'pact', 'bin', broker_client_exe()))

MESSAGE_PATH = normpath(join(
    dirname(__file__), 'bin', 'pact', 'bin', message_exe()))

MOCK_SERVICE_PATH = normpath(join(
    dirname(__file__), 'bin', 'pact', 'bin', mock_service_exe()))

VERIFIER_PATH = normpath(join(
    dirname(__file__), 'bin', 'pact', 'bin', provider_verifier_exe()))



"""API for creating a contract and configuring the mock service."""
from __future__ import unicode_literals

import json
import os
from subprocess import Popen

from .broker import Broker
from .constants import MESSAGE_PATH
from .matchers import from_term


class MessagePact(Broker):
    """
    Represents a contract between a consumer and provider using messages.

    Provides Python context handler to perform tests on a Python consumer. For example:

    >>> from pact import MessageConsumer, Provider
    >>> pact = MessageConsumer('MyMessageConsumer').has_pact_with(Provider('provider'))
    >>> (pact
    ... .given({"name": "Test provider"}])
    ... .expects_to_receive('Test description')
    ... .with_content({'name': 'John', 'document_name': 'sample_document.doc'})
    ... .with_metadata({'contentType': 'application/json'}))
    >>> with pact:
    ...   handler(event, context)
    """

    MANDATORY_FIELDS = {"providerStates", "description", "contents", "metaData"}

    def __init__(
        self,
        consumer,
        provider,
        publish_to_broker=False,
        broker_base_url=None,
        broker_username=None,
        broker_password=None,
        broker_token=None,
        pact_dir=None,
        version='3.0.0',
        file_write_mode="merge",
    ):
        """
        Create a Pact instance using messages.

        :param consumer: A consumer for this contract that uses messages.
        :type consumer: pact.MessageConsumer
        :param provider: The generic provider for this contract.
        :type provider: pact.Provider
        :param publish_to_broker: Flag to control automatic publishing of
            pacts to a pact broker. Defaults to False.
        :type publish_to_broker: bool
        :param broker_base_url: URL of the pact broker that pacts will be
            published to. Can also be supplied through the PACT_BROKER_BASE_URL
            environment variable. Defaults to None.
        :type broker_base_url: str
        :param broker_username: Username to use when connecting to the pact
            broker if authentication is required. Can also be supplied through
            the PACT_BROKER_USERNAME environment variable. Defaults to None.
        :type broker_username: str
        :param broker_password: Password to use when connecting to the pact
            broker if authentication is required. Strongly recommend supplying
            this value through the PACT_BROKER_PASSWORD environment variable
            instead. Defaults to None.
        :type broker_password: str
        :param broker_token: Authentication token to use when connecting to
            the pact broker. Strongly recommend supplying this value through
            the PACT_BROKER_TOKEN environment variable instead.
            Defaults to None.
        :type broker_token: str
        :param pact_dir: Directory where the resulting pact files will be
            written. Defaults to the current directory.
        :type pact_dir: str
        :param version: The Pact Specification version to use, defaults to
            '3.0.0'.
        :type version: str
        :param file_write_mode: `overwrite` or `merge`. Use `merge` when
            running multiple mock service instances in parallel for the same
            consumer/provider pair. Ensure the pact file is deleted before
            running tests when using this option so that interactions deleted
            from the code are not maintained in the file. Defaults to
            `merge`.
        :type file_write_mode: str
        """
        super().__init__(
            broker_base_url, broker_username, broker_password, broker_token
        )

        self.consumer = consumer
        self.file_write_mode = file_write_mode
        self.pact_dir = pact_dir or os.getcwd()
        self.provider = provider
        self.publish_to_broker = publish_to_broker
        self.version = version
        self._process = None
        self._messages = []
        self._message_process = None

    def given(self, provider_states):
        """
        Define the provider state for this pact.

        When the provider verifies this contract, they will use this field to
        setup pre-defined data that will satisfy the response expectations.

        :param provider_state: The short sentence that is unique to describe
            the provider state for this contract.
        :type provider_state: basestring
        :rtype: Pact
        """
        self._insert_message_if_complete()

        state = [{"name": "{}".format(provider_states)}]
        self._messages[0]['providerStates'] = state
        return self

    def with_metadata(self, metadata):
        """
        Define metadata attached to the message.

        :param metadata: dictionary of metadata attached to the message.
        :type metadata: dict or None
        :rtype: Pact
        """
        self._insert_message_if_complete()
        self._messages[0]['metaData'] = metadata
        return self

    def with_content(self, contents):
        """
        Define message content (event) that will be use in the message.

        :param contents: dictionary of dictionary used in the message.
        :type metadata: dict
        :rtype: Pact
        """
        self._insert_message_if_complete()
        self._messages[0]['contents'] = from_term(contents)
        return self

    def expects_to_receive(self, description):
        """
        Define the name of this contract (Same as upon_receiving in http pact implementation).

        :param scenario: A unique name for this contract.
        :type scenario: basestring
        :rtype: Pact
        """
        self._insert_message_if_complete()
        self._messages[0]['description'] = description
        return self

    def write_to_pact_file(self):
        """
        Create a pact file based on provided attributes in DSL.

        Return 0 if success, 1 otherwise.

        :rtype: int
        """
        command = [
            MESSAGE_PATH,
            "update",
            json.dumps(self._messages[0]),
            "--pact-dir", self.pact_dir,
            f"--pact-specification-version={self.version}",
            "--consumer", f"{self.consumer.name}",
            "--provider", f"{self.provider.name}",
        ]

        self._message_process = Popen(command)

    def _insert_message_if_complete(self):
        """
        Insert a new message if current message is complete.

        An interaction is complete if it has all the mandatory fields.
        If there are no message, a new message will be added.
        :rtype: None
        """
        if not self._messages:
            self._messages.append({})
        elif all(field in self._messages[0] for field in self.MANDATORY_FIELDS):
            self._messages.insert(0, {})

    def __enter__(self):
        """Enter a Python context. This function is required for context manager to work."""
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit a Python context.

        Calls pact-message to write out the contracts to disk.
        """
        if (exc_type, exc_val, exc_tb) != (None, None, None):
            return

        self.write_to_pact_file()

        if self.publish_to_broker:
            self.publish(
                self.consumer.name,
                self.consumer.version,
                pact_dir=self.pact_dir,
                tag_with_git_branch=self.consumer.tag_with_git_branch,
                consumer_tags=self.consumer.tags,
            )



"""Classes and methods to verify Contracts."""
import json

from pact.verify_wrapper import VerifyWrapper, path_exists, expand_directories

class Verifier(object):
    """A Pact Verifier."""

    def __init__(self, provider, provider_base_url, **kwargs):
        """Create a new Verifier.

        Args:
            provider ([String]): provider name
            provider_base_url ([String]): provider url

        """
        self.provider = provider
        self.provider_base_url = provider_base_url

    def __str__(self):
        """Return string representation.

        Returns:
            [String]: verifier description.

        """
        return 'Verifier for {} with url {}'.format(self.provider, self.provider_base_url)

    def validate_publish(self, **kwargs):
        """Validate publish has a version."""
        if ((kwargs.get('publish') is not None) and (kwargs.get('publish_version') is None)):
            # do something
            raise Exception()

    def verify_pacts(self, *pacts, enable_pending=False, include_wip_pacts_since=None, **kwargs):
        """Verify our pacts from the provider.

        Returns:
          success: True if no failures
          logs: some tbd output of logs

        """
        self.validate_publish(**kwargs)

        missing_files = [path for path in pacts if not path_exists(path)]
        if missing_files:
            raise Exception("Missing pact files {}".format(missing_files))

        pacts = expand_directories(pacts)
        missing_files = [path for path in pacts if not path_exists(path)]

        # rerun_command()  # env =

        options = self.extract_params(**kwargs)
        success, logs = VerifyWrapper().call_verify(*pacts,
                                                    provider=self.provider,
                                                    provider_base_url=self.provider_base_url,
                                                    enable_pending=enable_pending,
                                                    include_wip_pacts_since=include_wip_pacts_since,
                                                    **options)

        return success, logs

    def verify_with_broker(self, enable_pending=False, include_wip_pacts_since=None, **kwargs):
        """Use Broker to verify.

        Args:
            broker_username ([String]): broker username
            broker_password ([String]): broker password
            broker_url ([String]): url of broker
            enable_pending ([Boolean])
            include_wip_pacts_since ([String])
            publish_version ([String])

        """
        broker_username = kwargs.get('broker_username', None)
        broker_password = kwargs.get('broker_password', None)
        broker_url = kwargs.get('broker_url', None)
        broker_token = kwargs.get('broker_token', None)

        options = {
            'broker_password': broker_password,
            'broker_username': broker_username,
            'broker_token': broker_token,
            'broker_url': broker_url
        }
        options.update(self.extract_params(**kwargs))

        success, logs = VerifyWrapper().call_verify(provider=self.provider,
                                                    provider_base_url=self.provider_base_url,
                                                    enable_pending=enable_pending,
                                                    include_wip_pacts_since=include_wip_pacts_since,
                                                    **options)
        return success, logs

    def extract_params(self, **kwargs):
        """Extract params."""
        log_dir = kwargs.get('log_dir', None)
        log_level = kwargs.get('log_level', 'INFO')
        headers = kwargs.get('headers', [])
        timeout = kwargs.get('timeout', None)
        consumer_tags = kwargs.get('consumer_tags', [])
        provider_tags = kwargs.get('provider_tags', [])
        states_setup_url = kwargs.get('provider_states_setup_url', None)
        verbose = kwargs.get('verbose', False)
        provider_app_version = kwargs.get('publish_version', None)
        publish_verification_results = kwargs.get('publish_verification_results', None)
        raw_consumer_selectors = kwargs.get('consumer_version_selectors', [])
        consumer_selectors = self._build_consumer_selectors(raw_consumer_selectors)

        options = {
            'log_dir': log_dir,
            'log_level': log_level,
            'provider_app_version': provider_app_version,
            'custom_provider_headers': list(headers),
            'timeout': timeout,
            'consumer_tags': list(consumer_tags),
            'provider_tags': list(provider_tags),
            'provider_states_setup_url': states_setup_url,
            'verbose': verbose,
            'provider_app_version': provider_app_version,
            'consumer_selectors': consumer_selectors,
            'publish_verification_results': publish_verification_results
        }
        return self.filter_empty_options(**options)

    def _build_consumer_selectors(self, consumer_selectors):
        """
        Build the consumer_selectors list.

        Turn each dict in the consumer_selectors list into a string with a
        json object, as expected by VerifyWrapper.
        """
        return [json.dumps(selector) for selector in consumer_selectors]

    def filter_empty_options(self, **kwargs):
        """Filter out empty options."""
        kwargs = dict(filter(lambda item: item[1] is not None, kwargs.items()))
        kwargs = dict(filter(lambda item: item[1] != '', kwargs.items()))
        kwargs = dict(filter(lambda item: self.is_empty_list(item), kwargs.items()))
        return kwargs

    def is_empty_list(self, item):
        """Util for is empty lists."""
        return (not isinstance(item[1], list)) or (len(item[1]) != 0)



"""Methods to verify previously created pacts."""
import sys

from pact.verify_wrapper import path_exists, expand_directories, VerifyWrapper

import click


@click.command()
@click.argument('pacts', nargs=-1)
@click.option(
    'base_url', '--provider-base-url',
    help='Base URL of the provider to verify against.',
    required=True)
@click.option(
    'pact_url', '--pact-url',
    help='DEPRECATED: specify pacts as arguments instead.\n'
         'The URI of the pact to verify.'
         ' Can be an HTTP URI, a local file or directory path. '
         ' It can be specified multiple times to verify several pacts.',
    multiple=True)  # Remove in major version 1.0.0
@click.option(
    'pact_urls', '--pact-urls',
    default=[],
    help='DEPRECATED: specify pacts as arguments instead.\n'
         'The URI(s) of the pact to verify.'
         ' Can be an HTTP URI(s) or local file path(s).'
         ' Provide multiple URI separated by a comma.',
    multiple=True)  # Remove in major version 1.0.0
@click.option(
    'states_url', '--provider-states-url',
    help='DEPRECATED: URL to fetch the provider states for'
         ' the given provider API.')  # Remove in major version 1.0.0
@click.option(
    'states_setup_url', '--provider-states-setup-url',
    help='URL to send POST requests to setup a given provider state.')
@click.option(
    'username', '--pact-broker-username',
    envvar='PACT_BROKER_USERNAME',
    help='Username for Pact Broker basic authentication. Can also be specified'
         ' via the environment variable PACT_BROKER_USERNAME.')
@click.option(
    'broker_base_url', '--pact-broker-url',
    default='',
    envvar='PACT_BROKER_BASE_URL',
    help='Base URl for the Pact Broker instance to publish pacts to. Can also be specified'
         ' via the environment variable PACT_BROKER_BASE_URL.')
@click.option(
    'consumer_version_tag', '--consumer-version-tag',
    default=[],
    multiple=True,
    help='Retrieve the latest pacts with this consumer version tag. '
         'Used in conjunction with --provider. May be specified multiple times.')
@click.option(
    'consumer_version_selector', '--consumer-version-selector',
    default=[],
    multiple=True,
    help='Retrieve the latest pacts with this consumer version selector. '
         'Used in conjunction with --provider. May be specified multiple times.')
@click.option(
    'provider_version_tag', '--provider-version-tag',
    default=[],
    multiple=True,
    help='Tag to apply to the provider application version. '
         'May be specified multiple times.')
@click.option(
    'password', '--pact-broker-password',
    envvar='PACT_BROKER_PASSWORD',
    help='Password for Pact Broker basic authentication. Can also be specified'
         ' via the environment variable PACT_BROKER_PASSWORD.')
@click.option(
    'token', '--pact-broker-token',
    envvar='PACT_BROKER_TOKEN',
    help='Bearer token for Pact Broker authentication. Can also be specified'
         ' via the environment variable PACT_BROKER_TOKEN.')
@click.option(
    'provider', '--provider',
    default='',
    help='Retrieve the latest pacts for this provider.')
@click.option(
    'headers', '--custom-provider-header',
    envvar='CUSTOM_PROVIDER_HEADER',
    multiple=True,
    help='Header to add to provider state set up and '
         'pact verification requests. '
         'eg \'Authorization: Basic cGFjdDpwYWN0\'. '
         'May be specified multiple times.')
@click.option(
    'timeout', '-t', '--timeout',
    default=30,
    help='The duration in seconds we should wait to confirm that the verification'
         ' process was successful. Defaults to 30.',
    type=int)
@click.option(
    'provider_app_version', '-a', '--provider-app-version',
    help='The provider application version. '
         'Required for publishing verification results.')
@click.option(
    'publish_verification_results', '-r', '--publish-verification-results',
    default=False,
    help='Publish verification results to the broker.',
    is_flag=True)
@click.option(
    '--verbose/--no-verbose',
    default=False,
    help='Toggle verbose logging, defaults to False.')
@click.option(
    'log_dir', '--log-dir',
    help='The directory for the pact.log file.')
@click.option(
    'log_level', '--log-level',
    help='The logging level.')
@click.option(
    'enable_pending', '--enable-pending',
    default=False,
    help='Allow pacts which are in pending state to be verified without causing the '
         'overall task to fail. For more information, see https://pact.io/pending',
    is_flag=True)
@click.option(
    'include_wip_pacts_since', '--include-wip-pacts-since',
    default=None,
    help='Automatically include the pending pacts in the verification step. '
         'For more information, see https://docs.pact.io/pact_broker/advanced_topics/wip_pacts/',)
def main(pacts, base_url, pact_url, pact_urls, states_url, states_setup_url,
         username, broker_base_url, consumer_version_tag, consumer_version_selector,
         provider_version_tag, password, token, provider, headers, timeout,
         provider_app_version, publish_verification_results, verbose, log_dir,
         log_level, enable_pending, include_wip_pacts_since):
    """
    Verify one or more contracts against a provider service.

    Minimal example:
        pact-verifier --provider-base-url=http://localhost:8080 ./pacts
    """  # NOQA
    error = click.style('Error:', fg='red')
    warning = click.style('Warning:', fg='yellow')
    all_pact_urls = list(pacts) + list(pact_url)

    for urls in pact_urls:  # Remove in major version 1.0.0
        all_pact_urls.extend(p for p in urls.split(',') if p)

    if len(pact_urls) > 1:
        click.echo(
            warning
            + ' Multiple --pact-urls arguments are deprecated. '
            + 'Please provide a comma separated list of pacts to --pact-urls, '
            + 'or multiple --pact-url arguments.')

    if not all_pact_urls and broker_not_provided(broker_base_url, provider):
        click.echo(
            error
            + ' You must supply at least one pact file or directory '
            + 'to verify OR a Pact Broker and Provider.')
        raise click.Abort()

    all_pact_urls = expand_directories(all_pact_urls)

    missing_files = [path for path in all_pact_urls if not path_exists(path)]
    if missing_files:
        click.echo(
            error
            + ' The following Pact files could not be found:\n'
            + '\n'.join(missing_files))
        raise click.Abort()

    if publish_verification_results:
        validate_publish(error, provider_app_version)

    options = {
        'broker_password': password,
        'broker_username': username,
        'broker_token': token,
        'broker_url': broker_base_url,
        'log_dir': log_dir,
        'log_level': log_level,
        'provider_app_version': provider_app_version,
        'custom_provider_headers': list(headers),
        'publish_verification_results': publish_verification_results,
        'timeout': timeout,
        'verbose': verbose,
        'consumer_tags': list(consumer_version_tag),
        'consumer_selectors': list(consumer_version_selector),
        'provider_tags': list(provider_version_tag),
        'provider_states_setup_url': states_setup_url,
    }

    options = dict(filter(lambda item: item[1] is not None, options.items()))
    options = dict(filter(lambda item: item[1] != '', options.items()))
    options = dict(filter(lambda item: is_empty_list(item), options.items()))

    success, logs = VerifyWrapper().call_verify(*all_pact_urls,
                                                provider=provider,
                                                provider_base_url=base_url,
                                                enable_pending=enable_pending,
                                                include_wip_pacts_since=include_wip_pacts_since,
                                                **options)
    sys.exit(success)


def validate_publish(error, provider_app_version):
    """Publish results to broker."""
    if not provider_app_version:
        click.echo(
            error
            + 'Provider application version is required '
            + 'to publish verification results to broker'
        )
        raise click.Abort()


def broker_not_provided(broker_base_url, provider):
    """Check if broker not provided."""
    return (broker_base_url == '' or provider == '')


def is_empty_list(item):
    """Util for is empty lists."""
    return (not isinstance(item[1], list)) or (len(item[1]) != 0)


if __name__ == '__main__':
    sys.exit(main())



"""Python methods for interactive with a Pact Mock Service."""
from .broker import Broker
from .consumer import Consumer
from .matchers import EachLike, Like, SomethingLike, Term, Format
from .message_consumer import MessageConsumer
from .message_pact import MessagePact
from .message_provider import MessageProvider
from .pact import Pact
from .provider import Provider
from .verifier import Verifier

from .__version__ import __version__  # noqa: F401

__all__ = ('Broker', 'Consumer', 'EachLike', 'Like',
           'MessageConsumer', 'MessagePact', 'MessageProvider',
           'Pact', 'Provider', 'SomethingLike', 'Term', 'Format', 'Verifier')



"""API for creating a contract and configuring the mock service."""
from __future__ import unicode_literals

import fnmatch
import os
from subprocess import Popen

from .constants import BROKER_CLIENT_PATH

import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

class Broker():
    """PactBroker helper functions."""

    def __init__(self, broker_base_url=None, broker_username=None, broker_password=None, broker_token=None):
        """
        Create a Broker instance.

        :param broker_base_url: URL of the pact broker that pacts will be
            published to. Can also be supplied through the PACT_BROKER_BASE_URL
            environment variable. Defaults to None.
        :type broker_base_url: str
        :param broker_username: Username to use when connecting to the pact
            broker if authentication is required. Can also be supplied through
            the PACT_BROKER_USERNAME environment variable. Defaults to None.
        :type broker_username: str
        :param broker_password: Password to use when connecting to the pact
            broker if authentication is required. Strongly recommend supplying
            this value through the PACT_BROKER_PASSWORD environment variable
            instead. Defaults to None.
        :type broker_password: str
        :param broker_token: Authentication token to use when connecting to
            the pact broker. Strongly recommend supplying this value through
            the PACT_BROKER_TOKEN environment variable instead.
            Defaults to None.
        """
        self.broker_base_url = broker_base_url
        self.broker_username = broker_username
        self.broker_password = broker_password
        self.broker_token = broker_token

    def _get_broker_base_url(self):
        return self.broker_base_url or os.environ["PACT_BROKER_BASE_URL"]

    @staticmethod
    def _normalize_consumer_name(name):
        return name.lower().replace(' ', '_')

    def publish(self, consumer_name, version, pact_dir=None,
                tag_with_git_branch=None, consumer_tags=None):
        """Publish the generated pact files to the specified pact broker."""
        if self.broker_base_url is None \
                and "PACT_BROKER_BASE_URL" not in os.environ:
            raise RuntimeError("No pact broker URL specified. "
                               + "Did you expect the PACT_BROKER_BASE_URL "
                               + "environment variable to be set?")

        pact_files = fnmatch.filter(
            os.listdir(pact_dir),
            self._normalize_consumer_name(consumer_name) + '*.json'
        )
        pact_files = list(map(lambda pact_file: f'{pact_dir}/{pact_file}', pact_files))
        command = [
            BROKER_CLIENT_PATH,
            'publish',
            '--consumer-app-version={}'.format(version)]

        command.append('--broker-base-url={}'.format(self._get_broker_base_url()))

        if self.broker_username is not None:
            command.append('--broker-username={}'.format(self.broker_username))
        if self.broker_password is not None:
            command.append('--broker-password={}'.format(self.broker_password))
        if self.broker_token is not None:
            command.append('--broker-token={}'.format(self.broker_token))

        command.extend(pact_files)

        if tag_with_git_branch:
            command.append('--tag-with-git-branch')

        if consumer_tags is not None:
            for tag in consumer_tags:
                command.extend(['-t', tag])

        log.debug(f"PactBroker publish command: {command}")

        publish_process = Popen(command)
        publish_process.wait()
        if publish_process.returncode != 0:
            url = self._get_broker_base_url()
            raise RuntimeError(
                f"There was an error while publishing to the pact broker at {url}.")



"""Wrapper to verify previously created pacts."""

from pact.constants import VERIFIER_PATH
import sys
import os
import platform

import subprocess
from os.path import isdir, join, isfile
from os import listdir

def capture_logs(process, verbose):
    """Capture logs from ruby process."""
    result = ''
    for line in process.stdout:
        result = result + line + '\n'

    return result


def path_exists(path):
    """
    Determine if a particular path exists.

    Can be provided a URL or local path. URLs always result in a True. Local
    paths are True only if a file exists at that location.

    :param path: The path to check.
    :type path: str
    :return: True if the path exists and is a file, otherwise False.
    :rtype: bool
    """
    if path.startswith('http://') or path.startswith('https://'):
        return True

    return isfile(path)

def sanitize_logs(process, verbose):
    """
    Print the logs from a process while removing Ruby stack traces.

    :param process: The Ruby pact verifier process.
    :type process: subprocess.Popen
    :param verbose: Flag to toggle more verbose logging.
    :type verbose: bool
    :rtype: None
    """
    for line in process.stdout:
        if (not verbose and line.lstrip().startswith('#')
                and ('vendor/ruby' in line or 'pact-provider-verifier.rb' in line)):
            continue
        else:
            sys.stdout.write(line)

def expand_directories(paths):
    """
    Iterate over paths and expand any that are directories into file paths.

    :param paths: A list of file paths to expand.
    :type paths: list
    :return: A list of file paths with any directory paths replaced with the
        JSON files in those directories.
    :rtype: list
    """
    paths_ = []
    for path in paths:
        if path.startswith('http://') or path.startswith('https://'):
            paths_.append(path)
        elif isdir(path):
            paths_.extend(
                [join(path, p) for p in listdir(path) if p.endswith('.json')])
        else:
            paths_.append(path)

    # Ruby pact verifier expects forward slashes regardless of OS
    return [p.replace('\\', '/') for p in paths_]


def rerun_command():
    """
    Create a rerun command template for failed interactions.

    :rtype: str
    """
    is_windows = 'windows' in platform.platform().lower()
    command = ''
    if is_windows:
        command = (
            'cmd.exe /v /c "'
            'set PACT_DESCRIPTION=<PACT_DESCRIPTION>'
            '& set PACT_PROVIDER_STATE=<PACT_PROVIDER_STATE>'
            '& {command}'
            ' & set PACT_DESCRIPTION='
            ' & set PACT_PROVIDER_STATE="'.format(command=' '.join(sys.argv)))
    else:
        command = ("PACT_DESCRIPTION='<PACT_DESCRIPTION>'"
                   " PACT_PROVIDER_STATE='<PACT_PROVIDER_STATE>'"
                   " {command}".format(command=' '.join(sys.argv)))

    env = os.environ.copy()
    env['PACT_INTERACTION_RERUN_COMMAND'] = command
    return env

class PactException(Exception):
    """PactException when input isn't valid.

    Args:
        Exception ([type]): [description]

    Raises:
        KeyError: [description]
        Exception: [description]

    Returns:
        [type]: [description]

    """

    def __init__(self, *args, **kwargs):
        """Create wrapper."""
        super().__init__(*args, **kwargs)
        self.message = args[0]

class VerifyWrapper(object):
    """A Pact Verifier Wrapper."""

    def _broker_present(self, **kwargs):
        if kwargs.get('broker_url') is None:
            return False
        return True

    def _validate_input(self, pacts, **kwargs):
        if len(pacts) == 0 and not self._broker_present(**kwargs):
            raise PactException('Pact urls or Pact broker required')

    def call_verify(
            self, *pacts, provider_base_url, provider, enable_pending=False,
            include_wip_pacts_since=None, **kwargs
    ):
        """Call verify method."""
        verbose = kwargs.get('verbose', False)

        self._validate_input(pacts, **kwargs)

        provider_app_version = kwargs.get('provider_app_version')
        options = {
            '--provider-base-url': provider_base_url,
            '--provider': provider,
            '--broker-username': kwargs.get('broker_username', None),
            '--broker-password': kwargs.get('broker_password', None),
            '--broker-token': kwargs.get('broker_token', None),
            '--pact-broker-base-url': kwargs.get('broker_url', None),
            '--provider-states-setup-url': kwargs.get('provider_states_setup_url'),
            '--log-dir': kwargs.get('log_dir'),
            '--log-level': kwargs.get('log_level')
        }

        command = [VERIFIER_PATH]
        all_pact_urls = expand_directories(list(pacts))

        command.extend(all_pact_urls)
        command.extend(['{}={}'.format(k, v) for k, v in options.items() if v])

        if(provider_app_version):
            command.extend(["--provider-app-version",
                            provider_app_version])

        if(kwargs.get('publish_verification_results', False) is True):
            command.extend(['--publish-verification-results'])

        if(kwargs.get('verbose', False) is True):
            command.extend(['--verbose'])
        if enable_pending:
            command.append('--enable-pending')

        if include_wip_pacts_since:
            command.extend(['--include-wip-pacts-since={}'.format(include_wip_pacts_since)])

        headers = kwargs.get('custom_provider_headers', [])
        for header in headers:
            command.extend(['{}={}'.format('--custom-provider-header', header)])
        for tag in kwargs.get('consumer_tags', []):
            command.extend(["--consumer-version-tag={}".format(tag)])
        for tag in kwargs.get('consumer_selectors', []):
            command.extend(["--consumer-version-selector={}".format(tag)])
        for tag in kwargs.get('provider_tags', []):
            command.extend(["--provider-version-tag={}".format(tag)])

        env = rerun_command()
        result = subprocess.Popen(command, bufsize=1, env=env, stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT, universal_newlines=True)

        sanitize_logs(result, verbose)
        result.wait()
        logs = capture_logs(result, verbose)

        return result.returncode, logs

    def publish_results(self, provider_app_version, command):
        """Publish results to broker."""
        if not provider_app_version:
            # todo implement
            raise Exception('todo')

        command.extend(["--provider-app-version",
                       provider_app_version,
                       "--publish-verification-results"])



"""Http Proxy to be used as provider url in verifier."""
from fastapi import FastAPI, status, Request, HTTPException
import uvicorn as uvicorn
import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

app = FastAPI()
PROXY_PORT = 1234
UVICORN_LOGGING_LEVEL = "error"
items = {
    "states": None
}


def _match_states(payload):
    """Match states in payload against stored message handlers."""
    log.debug(f'Find handler from payload: {payload}')
    handlers = items["states"]
    states = handlers['messageHandlers']
    log.debug(f'Setup states: {handlers}')
    provider_states = payload['providerStates']

    for state in provider_states:
        matching_state = state['name']
        if matching_state in states:
            return states[matching_state]
    raise HTTPException(status_code=500, detail='No matched handler.')


@app.post("/")
async def root(request: Request):
    """Match states with provided message handlers."""
    payload = await request.json()
    message = _match_states(payload)
    return {'contents': message}


@app.get('/ping', status_code=status.HTTP_200_OK)
def ping():
    """Check whether the server is available before setting up states."""
    return {"ping": "pong"}


@app.post("/setup", status_code=status.HTTP_201_CREATED)
async def setup(request: Request):
    """Endpoint to setup states.

    Use localstack to store payload.
    """
    payload = await request.json()
    items["states"] = payload
    return items["states"]


def run_proxy():
    """Rub HTTP Proxy."""
    uvicorn.run("pact.http_proxy:app", port=PROXY_PORT, log_level=UVICORN_LOGGING_LEVEL)



"""pact-python PyPI Package."""

import os
import platform
import sys
import tarfile

from zipfile import ZipFile

from setuptools import setup
from setuptools.command.develop import develop
from setuptools.command.install import install


IS_64 = sys.maxsize > 2 ** 32
PACT_STANDALONE_VERSION = '1.88.51'


here = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(here, "pact", "__version__.py")) as f:
    exec(f.read(), about)


class PactPythonDevelopCommand(develop):
    """
    Custom develop mode installer for pact-python.

    When the package is installed using `python setup.py develop` or
    `pip install -e` it will download and unpack the appropriate Pact
    mock service and provider verifier.
    """

    def run(self):
        """Install ruby command."""
        develop.run(self)
        bin_path = os.path.join(os.path.dirname(__file__), 'pact', 'bin')
        if not os.path.exists(bin_path):
            os.mkdir(bin_path)

        install_ruby_app(bin_path)


class PactPythonInstallCommand(install):
    """
    Custom installer for pact-python.

    Installs the Python package and unpacks the platform appropriate version
    of the Ruby mock service and provider verifier.
    """

    def run(self):
        """Install python binary."""
        install.run(self)
        bin_path = os.path.join(self.install_lib, 'pact', 'bin')
        os.mkdir(bin_path)
        install_ruby_app(bin_path)


def install_ruby_app(bin_path):
    """
    Download a Ruby application and install it for use.

    :param bin_path: The path where binaries should be installed.
    """
    target_platform = platform.platform().lower()
    uri = ('https://github.com/pact-foundation/pact-ruby-standalone/releases'
           '/download/v{version}/pact-{version}-{suffix}')

    if 'darwin' in target_platform or 'macos' in target_platform:
        suffix = 'osx.tar.gz'
    elif 'linux' in target_platform and IS_64:
        suffix = 'linux-x86_64.tar.gz'
    elif 'linux' in target_platform:
        suffix = 'linux-x86.tar.gz'
    elif 'windows' in target_platform:
        suffix = 'win32.zip'
    else:
        msg = ('Unfortunately, {} is not a supported platform. Only Linux,'
               ' Windows, and OSX are currently supported.').format(
            platform.platform())
        raise Exception(msg)

    if sys.version_info.major == 2:
        from urllib import urlopen
    else:
        from urllib.request import urlopen

    path = os.path.join(bin_path, suffix)
    resp = urlopen(uri.format(version=PACT_STANDALONE_VERSION, suffix=suffix))
    with open(path, 'wb') as f:
        if resp.code == 200:
            f.write(resp.read())
        else:
            raise RuntimeError(
                'Received HTTP {} when downloading {}'.format(
                    resp.code, resp.url))

    if 'windows' in platform.platform().lower():
        with ZipFile(path) as f:
            f.extractall(bin_path)
    else:
        with tarfile.open(path) as f:
            f.extractall(bin_path)


def read(filename):
    """Read file contents."""
    path = os.path.realpath(os.path.join(os.path.dirname(__file__), filename))
    with open(path, 'rb') as f:
        return f.read().decode('utf-8')


dependencies = [
    'click>=2.0.0',
    'psutil>=2.0.0',
    'requests>=2.5.0',
    'six>=1.9.0',
    'fastapi==0.67.0',
    'urllib3>=1.26.5',
    'uvicorn==0.14.0'
]

if __name__ == '__main__':
    setup(
        cmdclass={
            'develop': PactPythonDevelopCommand,
            'install': PactPythonInstallCommand},
        name='pact-python',
        version=about['__version__'],
        description=(
            'Tools for creating and verifying consumer driven '
            'contracts using the Pact framework.'),
        long_description=read('README.md'),
        long_description_content_type='text/markdown',
        author='Matthew Balvanz',
        author_email='matthew.balvanz@workiva.com',
        url='https://github.com/pact-foundation/pact-python',
        entry_points='''
            [console_scripts]
            pact-verifier=pact.cli.verify:main
        ''',
        install_requires=dependencies,
        packages=['pact', 'pact.cli'],
        package_data={'pact': ['bin/*']},
        package_dir={'pact': 'pact'},
        license='MIT License')



"""API for creating a contract and configuring the mock service."""
from __future__ import unicode_literals

import os
import platform
from subprocess import Popen

import psutil
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3 import Retry

from .broker import Broker
from .constants import MOCK_SERVICE_PATH
from .matchers import from_term


class Pact(Broker):
    """
    Represents a contract between a consumer and provider.

    Provides Python context handlers to configure the Pact mock service to
    perform tests on a Python consumer. For example:

    >>> from pact import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact.given('the echo service is available')
    ...  .upon_receiving('a request is made to the echo service')
    ...  .with_request('get', '/echo', query={'text': 'Hello!'})
    ...  .will_respond_with(200, body='Hello!'))
    >>> with pact:
    ...   requests.get(pact.uri + '/echo?text=Hello!')

    The GET request is made to the mock service, which will verify that it
    was a GET to /echo with a query string with a key named `text` and its
    value is `Hello!`. If the request does not match an error is raised, if it
    does match the defined interaction, it will respond with the text `Hello!`.
    """

    HEADERS = {'X-Pact-Mock-Service': 'true'}

    MANDATORY_FIELDS = {'response', 'description', 'request'}

    def __init__(
        self,
        consumer,
        provider,
        host_name='localhost',
        port=1234,
        log_dir=None,
        ssl=False,
        sslcert=None,
        sslkey=None,
        cors=False,
        publish_to_broker=False,
        broker_base_url=None,
        broker_username=None,
        broker_password=None,
        broker_token=None,
        pact_dir=None,
        specification_version='2.0.0',
        file_write_mode='overwrite',
    ):
        """
        Create a Pact instance.

        :param consumer: The consumer for this contract.
        :type consumer: pact.Consumer
        :param provider: The provider for this contract.
        :type provider: pact.Provider
        :param host_name: The host name where the mock service is running.
        :type host_name: str
        :param port: The port number where the mock service is running.
        :type port: int
        :param log_dir: The directory where logs should be written. Defaults to
            the current directory.
        :type log_dir: str
        :param ssl: Flag to control the use of a self-signed SSL cert to run
            the server over HTTPS , defaults to False.
        :type ssl: bool
        :param sslcert: Path to a custom self-signed SSL cert file, 'ssl'
            option must be set to True to use this option. Defaults to None.
        :type sslcert: str
        :param sslkey: Path to a custom key and self-signed SSL cert key file,
            'ssl' option must be set to True to use this option.
            Defaults to None.
        :type sslkey: str
        :param cors: Allow CORS OPTION requests to be accepted,
            defaults to False.
        :type cors: bool
        :param publish_to_broker: Flag to control automatic publishing of
            pacts to a pact broker. Defaults to False.
        :type publish_to_broker: bool
        :param broker_base_url: URL of the pact broker that pacts will be
            published to. Can also be supplied through the PACT_BROKER_BASE_URL
            environment variable. Defaults to None.
        :type broker_base_url: str
        :param broker_username: Username to use when connecting to the pact
            broker if authentication is required. Can also be supplied through
            the PACT_BROKER_USERNAME environment variable. Defaults to None.
        :type broker_username: str
        :param broker_password: Password to use when connecting to the pact
            broker if authentication is required. Strongly recommend supplying
            this value through the PACT_BROKER_PASSWORD environment variable
            instead. Defaults to None.
        :type broker_password: str
        :param broker_token: Authentication token to use when connecting to
            the pact broker. Strongly recommend supplying this value through
            the PACT_BROKER_TOKEN environment variable instead.
            Defaults to None.
        :type broker_token: str
        :param pact_dir: Directory where the resulting pact files will be
            written. Defaults to the current directory.
        :type pact_dir: str
        :param specification_version: The Pact Specification version to use, defaults to
            '2.0.0'.
        :type version: str of the consumer version.
        :param file_write_mode: `overwrite` or `merge`. Use `merge` when
            running multiple mock service instances in parallel for the same
            consumer/provider pair. Ensure the pact file is deleted before
            running tests when using this option so that interactions deleted
            from the code are not maintained in the file. Defaults to
            `overwrite`.
        :type file_write_mode: str
        """
        super().__init__(
            broker_base_url, broker_username, broker_password, broker_token
        )

        scheme = 'https' if ssl else 'http'
        self.uri = '{scheme}://{host_name}:{port}'.format(
            host_name=host_name, port=port, scheme=scheme)
        self.consumer = consumer
        self.cors = cors
        self.file_write_mode = file_write_mode
        self.host_name = host_name
        self.log_dir = log_dir or os.getcwd()
        self.pact_dir = pact_dir or os.getcwd()
        self.port = port
        self.provider = provider
        self.publish_to_broker = publish_to_broker
        self.ssl = ssl
        self.sslcert = sslcert
        self.sslkey = sslkey
        self.specification_version = specification_version
        self._interactions = []
        self._process = None

    def given(self, provider_state):
        """
        Define the provider state for this pact.

        When the provider verifies this contract, they will use this field to
        setup pre-defined data that will satisfy the response expectations.

        :param provider_state: The short sentence that is unique to describe
            the provider state for this contract.
        :type provider_state: basestring
        :rtype: Pact
        """
        self._insert_interaction_if_complete()
        self._interactions[0]['provider_state'] = provider_state
        return self

    def setup(self):
        """Configure the Mock Service to ready it for a test."""
        try:
            interactions_uri = f"{self.uri}/interactions"
            resp = requests.delete(
                interactions_uri, headers=self.HEADERS, verify=False
            )

            assert resp.status_code == 200, resp.text
            resp = requests.put(
                interactions_uri,
                headers=self.HEADERS,
                verify=False,
                json={"interactions": self._interactions},
            )

            assert resp.status_code == 200, resp.text
        except AssertionError:
            raise

    def start_service(self):
        """
        Start the external Mock Service.

        :raises RuntimeError: if there is a problem starting the mock service.
        """
        command = [
            MOCK_SERVICE_PATH,
            "service",
            f"--host={self.host_name}",
            f"--port={format(self.port)}",
            "--log", f"{self.log_dir}/pact-mock-service.log",
            "--pact-dir", self.pact_dir,
            "--pact-file-write-mode", self.file_write_mode,
            f"--pact-specification-version={self.specification_version}",
            "--consumer", self.consumer.name,
            "--provider", self.provider.name,
        ]

        if self.ssl:
            command.append('--ssl')
        if self.sslcert:
            command.extend(['--sslcert', self.sslcert])
        if self.sslkey:
            command.extend(['--sslkey', self.sslkey])

        self._process = Popen(command)
        self._wait_for_server_start()

    def stop_service(self):
        """Stop the external Mock Service."""
        is_windows = 'windows' in platform.platform().lower()
        if is_windows:
            # Send the signal to ruby.exe, not the *.bat process
            p = psutil.Process(self._process.pid)
            for child in p.children(recursive=True):
                child.terminate()
            p.wait()
            if psutil.pid_exists(self._process.pid):
                raise RuntimeError(
                    'There was an error when stopping the Pact mock service.')

        else:
            self._process.terminate()

            self._process.communicate()
            if self._process.returncode != 0:
                raise RuntimeError(
                    'There was an error when stopping the Pact mock service.'
                )
        if self.publish_to_broker:
            self.publish(
                self.consumer.name,
                self.consumer.version,
                tag_with_git_branch=self.consumer.tag_with_git_branch,
                consumer_tags=self.consumer.tags,
                pact_dir=self.pact_dir
            )

    def upon_receiving(self, scenario):
        """
        Define the name of this contract.

        :param scenario: A unique name for this contract.
        :type scenario: basestring
        :rtype: Pact
        """
        self._insert_interaction_if_complete()
        self._interactions[0]['description'] = scenario
        return self

    def verify(self):
        """
        Have the mock service verify all interactions occurred.

        Calls the mock service to verify that all interactions occurred as
        expected, and has it write out the contracts to disk.

        :raises AssertionError: When not all interactions are found.
        """
        self._interactions = []
        resp = requests.get(
            self.uri + "/interactions/verification", headers=self.HEADERS, verify=False
        )
        assert resp.status_code == 200, resp.text
        resp = requests.post(self.uri + "/pact", headers=self.HEADERS, verify=False)
        assert resp.status_code == 200, resp.text

    def with_request(self, method, path, body=None, headers=None, query=None):
        """
        Define the request that the client is expected to perform.

        :param method: The HTTP method.
        :type method: str
        :param path: The path portion of the URI the client will access.
        :type path: str, Matcher
        :param body: The request body, can be a string or an object that will
            serialize to JSON, like list or dict, defaults to None.
        :type body: list, dict or None
        :param headers: The headers the client is expected to include on with
            this request. Defaults to None.
        :type headers: dict or None
        :param query: The query options the client is expected to send. Can be
            a dict of keys and values, or a URL encoded string.
            Defaults to None.
        :type query: dict, basestring, or None
        :rtype: Pact
        """
        self._insert_interaction_if_complete()
        self._interactions[0]['request'] = Request(
            method, path, body=body, headers=headers, query=query
        ).json()
        return self

    def will_respond_with(self, status, headers=None, body=None):
        """
        Define the response the server is expected to create.

        :param status: The HTTP status code.
        :type status: int
        :param headers: All required headers. Defaults to None.
        :type headers: dict or None
        :param body: The response body, or a collection of Matcher objects to
            allow for pattern matching. Defaults to None.
        :type body: Matcher, dict, list, basestring, or None
        :rtype: Pact
        """
        self._insert_interaction_if_complete()
        self._interactions[0]['response'] = Response(
            status, headers=headers, body=body
        ).json()
        return self

    def _insert_interaction_if_complete(self):
        """
        Insert a new interaction if current interaction is complete.

        An interaction is complete if it has all the mandatory fields.
        If there are no interactions, a new interaction will be added.

        :rtype: None
        """
        if not self._interactions:
            self._interactions.append({})
        elif all(field in self._interactions[0] for field in self.MANDATORY_FIELDS):
            self._interactions.insert(0, {})

    def _wait_for_server_start(self):
        """
        Wait for the mock service to be ready for requests.

        :rtype: None
        :raises RuntimeError: If there is a problem starting the mock service.
        """
        s = requests.Session()
        retries = Retry(total=9, backoff_factor=0.1)
        http_mount = 'https://' if self.ssl else 'http://'
        s.mount(http_mount, HTTPAdapter(max_retries=retries))

        resp = s.get(self.uri, headers=self.HEADERS, verify=False)
        if resp.status_code != 200:
            self._process.terminate()
            self._process.communicate()
            raise RuntimeError(
                'There was a problem starting the mock service: %s', resp.text
            )

    def __enter__(self):
        """
        Enter a Python context.

        Sets up the mock service to expect the client requests.
        """
        self.setup()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit a Python context.

        Calls the mock service to verify that all interactions occurred as
        expected, and has it write out the contracts to disk.
        """
        if (exc_type, exc_val, exc_tb) != (None, None, None):
            return

        self.verify()


class FromTerms(object):
    """Base class for objects built from a collection of Matchers."""

    def json(self):
        """Convert the object to a JSON version of the mock service."""
        raise NotImplementedError


class Request(FromTerms):
    """Represents an HTTP request and supports Matchers on its properties."""

    def __init__(self, method, path, body=None, headers=None, query=''):
        """
        Create a new instance of Request.

        :param method: The HTTP method that is expected.
        :type method: str
        :param path: The URI path that is expected on this request.
        :type path: str, Matcher
        :param body: The contents of the body of the expected request.
        :type body: str, dict, list
        :param headers: The headers of the expected request.
        :type headers: dict
        :param query: The URI query of the expected request.
        :type query: str or dict
        """
        self.method = method
        self.path = from_term(path)
        self.body = from_term(body)
        self.headers = from_term(headers)
        self.query = from_term(query)

    def json(self):
        """Convert the Request to a JSON version for the mock service."""
        request = {'method': self.method, 'path': self.path}

        if self.headers:
            request['headers'] = self.headers

        if self.body is not None:
            request['body'] = self.body

        if self.query:
            request['query'] = self.query

        return request


class Response(FromTerms):
    """Represents an HTTP response and supports Matchers on its properties."""

    def __init__(self, status, headers=None, body=None):
        """
        Create a new Response.

        :param status: The expected HTTP status of the response.
        :type status: int
        :param headers: The expected headers of the response.
        :type headers: dict
        :param body: The expected body of the response.
        :type body: str, dict, or list
        """
        self.status = status
        self.body = from_term(body)
        self.headers = from_term(headers)

    def json(self):
        """Convert the Response to a JSON version for the mock service."""
        response = {'status': self.status}
        if self.body is not None:
            response['body'] = self.body

        if self.headers:
            response['headers'] = self.headers

        return response
